"""RDF diff with bnode handling command line utility."""

import logging
import sys

from rdflib import __version__ as rdflib_version, Graph
from rdflib.compare import isomorphic, to_isomorphic, graph_diff, to_canonical_graph
from rdflib.util import guess_format
from rdflib.term import BNode, URIRef

from .canonicalizer_with_memory import CanonicalizerWithMemory
from .bnode_mapper import nt_sorted, serialize_equalish, to_bnodes, from_bnodes_triple, from_bnodes


class RDiffB(object):
    """RDiffB class implements RDF diff functions with bnode mapping."""

    def __init__(self, bnodes=None):
        """Initialize, optionally with set of bnode patterns."""
        self.bnodes = bnodes if bnodes else []
        self._initialize()

    def _initialize(self):
        # Clear internal data
        self.filenames = []
        self.graphs = []
        self.mappings = []
        self.num_subs = 0
        # results
        self.in_both = None
        self.in_first = None
        self.in_second = None

    @property
    def num_diff(self):
        """Number of triples in difference."""
        return(len(self.in_first) + len(self.in_second))

    def load_graph(self, filename):
        """Load RDF from file to get graph."""
        fmt = guess_format(filename)
        logging.info("Reading %s as %s..." % (filename, fmt))
        graph = Graph().parse(filename, format=fmt)
        logging.debug("... got %d triples" % (len(graph)))
        return(graph)

    def map_patterns_and_canonicalize(self, graph):
        """Map nodes matching patterns in self.bnodes and canonicalize.

        Records the mappings applied, including following changes made
        during the canonicalization process. Returns the modified graph
        and the mapping.
        """
        mapping = dict()
        for pattern in self.bnodes:
            logging.debug("Looking for pattern %s" % (pattern))
            graph, subs, this_mapping = to_bnodes(graph, pattern)
            self.num_subs += subs
            mapping.update(this_mapping)
        logging.debug("Complete mapping: " + str(mapping))
        cwm = CanonicalizerWithMemory(graph)
        graph = cwm.canonical_graph()
        # Merge mapping from patterns with mapping from canonicalization
        mapping = dict((v, k) for k, v in mapping.items())  # invert: bnode -> original URI
        for (new_bnode, old_bnode) in cwm.bnode_map.items():
            if (old_bnode in mapping):
                # We mapped twice, make make from new_bnode -> original
                mapping[new_bnode] = mapping[old_bnode]
                del mapping[old_bnode]
            else:
                # Original was a bnode, just canonicalization mapping
                mapping[new_bnode] = old_bnode
        return(graph, mapping)

    def compare_files(self, filenames):
        """Compare graphs from two RDF files in the filenames list.

        Returns the number of triples that are different after matching
        up bnodes.
        """
        graphs = []
        for filename in filenames:
            graphs.append(self.load_graph(filename))
        return(self.compare_graphs(graphs, filenames))

    def compare_graphs(self, graphs, names=None):
        """Compare two Graph objects given in the graphs list.

        Returns the number of triples that are different after matching
        up bnodes.
        """
        self._initialize()
        if (names is None):
            self.filenames = ['graph1', 'graph2']
        else:
            self.filenames = names
        for g in graphs:
            (graph, mapping) = self.map_patterns_and_canonicalize(g)
            self.graphs.append(graph)
            self.mappings.append(mapping)
        self.in_both, self.in_first, self.in_second = \
            graph_diff(self.graphs[0], self.graphs[1])
        return(self.num_diff)

    def write_diff(self, fh, report_isomorphic=False):
        """Write diff results in human-readable form to fh."""
        num_same = len(self.in_both)
        pct_same = num_same * 200.0 / (self.num_diff + 2 * num_same)
        if (self.num_diff == 0):
            if (report_isomorphic):
                sub_note = ' after bnode substitutions' if (self.num_subs > 0) else ''
                fh.write("Graphs %s and %s are isomorphic%s\n" %
                         (self.filenames[0], self.filenames[1], sub_note))
        else:
            logging.info(
                "%d triples are shared by the two graphs (%.1f%%)" %
                (num_same, pct_same))
            logging.info(serialize_equalish(self.in_both, self.mappings))
            # Normal diff outout for < and >
            if (len(self.in_first) > 0):
                logging.debug("In first only:")
                fh.write(nt_sorted(from_bnodes(self.in_first, self.mappings[0]), '<  ') + "\n")
            if (len(self.in_second) > 0):
                logging.debug("In second only:")
                fh.write(nt_sorted(from_bnodes(self.in_second, self.mappings[1]), '>  ') + "\n")



#!/usr/bin/env python
"""
rdfdiffb.py: RDF diff with bnode handling command line utility.

Copyright 2016--2020

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License
"""

import logging
import optparse
import sys

from rdflib import __version__ as rdflib_version, Graph
from rdflib.compare import isomorphic, to_isomorphic, graph_diff, to_canonical_graph
from rdflib.util import guess_format
from rdflib.term import BNode, URIRef

from rdiffb import __version__ as rdiffb_version, RDiffB
from rdiffb.bnode_mapper import nt_sorted, serialize_equalish, from_bnodes


def main():
    """Command line client."""
    if (sys.version_info < (2, 7)):
        sys.exit("This program requires python version 2.7 or later")
    if (rdflib_version < "4.2.0"):
        sys.exit("This program requires rdflib >= 4.2.0 for correct "
                 "rdflib.compare using the RDGA1 algorithm")

    # Options and arguments
    p = optparse.OptionParser(description='rdfdiffb command line client',
                              usage='usage: %prog [options] uri_path local_path  (-h for help)',
                              version='%prog ' + rdiffb_version)

    p.add_option('--bnode', '-b', action='append', default=[],
                 help="add a regex for URIs that should be treated like bnodes "
                      "in diffs (repeatable)")
    p.add_option('--report-isomorphic-graphs', '-s', action='store_true',
                 help="report if the input RDF graphs are isomorphic (after any bnode substitution)")
    p.add_option('--verbose', '-v', action='store_true',
                 help="verbose, show additional informational messages")
    p.add_option('--debug', '-d', action='store_true',
                 help="very verbose output")

    (opt, args) = p.parse_args()

    # use --verbose & --debug to set up logging level
    level = logging.DEBUG if opt.debug else logging.INFO if opt.verbose else logging.WARN
    logging.basicConfig(level=level, format='%(message)s')

    if (len(args) != 2):
        sys.exit("Two file name arguments required")

    rdb = RDiffB(opt.bnode)
    num_diff = rdb.compare_files(args)
    rdb.write_diff(sys.stdout, opt.report_isomorphic_graphs)

    # Exit status: 0 if same, 1 if different
    sys.exit(1 if (num_diff > 0) else 0)


if __name__ == '__main__':
    main()



"""Create canonical graph and remember bnode id canonicalization.

This code subclasses rdflib.compare._TripleCanonicalizer in order
add storage for the bnode id changes, and to store each change.

WARNING - This code is critically dependent upon the internal
structure of the base class rdflib.compare._TripleCanonicalizer
and this is not a published interface --- IT MAY BREAK WITH A
NEW VERSION OF rdflib (created with v4.2.1 2016-09-09).
"""

from rdflib.graph import Graph, ConjunctiveGraph, ReadOnlyGraphAggregate
from rdflib.term import BNode, Node
from rdflib.compare import _TripleCanonicalizer


class CanonicalizerWithMemory(_TripleCanonicalizer):
    """Graph canonicalization class.

    Subclass of rdflib.compare._TripleCanonicalizer that adds
    storage of bnode id mappings during canonicalization.
    """

    def __init__(self, graph, **kwargs):
        """Initialize instance, add bnode_map."""
        super(CanonicalizerWithMemory, self).__init__(graph, **kwargs)
        self.bnode_map = dict()

    def _canonicalize_bnodes(self, triple, labels):
        """Assign new bnode ids based on canonical labels.

        WARNING - This method overrides the method of the same name
        in rdflib.compare._TripleCanonicalizer --- IT MAY BREAK WITH
        A NEW VERSION OF rdflib (created with v4.2.1).

        For each bnode id change, the map from new BNode to the
        old BNode is recorded in self.bnode_map.
        """
        for term in triple:
            if isinstance(term, BNode):
                # FIXME - it is a bit inefficient to do this here as we
                # FIXME - set the value in the bnode_map dict() every time
                # FIXME - the mapping is used. However, since this is where
                # FIXME - in the code the full id is generated it seems the
                # FIXME - cleanest place to put it. Otherwise we'd have to
                # FIXME - duplicate the id construction from label[..]
                new_bnode = BNode(value="cb%s" % labels[term])
                # print(str(term) + ' --> ' + str(new_bnode))
                self.bnode_map[new_bnode] = term
                yield new_bnode
            else:
                yield term

    def canonical_graph(self):
        """Create a canonical, read-only graph.

        Modification of rdflib.compare.to_canonical_graph that is instead
        made as a method of CanonicalizerWithMemory so that we can access
        the store of bnode relabelings made during the canonicalization.

        Creates a canonical, read-only graph where all bnode id's are based on
        deterministical SHA-256 checksums, correlated with the graph contents.

        As a side-effect, self.bnode_map will be populated with a mapping
        from new bnode ids to old bnode ids (see _canonicalize_bnodes).
        """
        graph = Graph()
        graph += self.canonical_triples()
        return ReadOnlyGraphAggregate([graph])



"""Functions for mapping to and from bnodes."""

import logging
from rdflib import Graph
from rdflib.term import BNode, URIRef
from rdflib.plugins.serializers.nt import _nt_row
import re


def nt_sorted(g, prefix=''):
    """Linewise sort ntriples serialization, with optional prefix.

    Does not include a trailing return character.
    """
    s = ''
    for line in sorted(g.serialize(format='nt').splitlines()):
        if line:
            s += '\n' if (s) else ''
            s += prefix + line.decode('utf-8')
    return s


def serialize_equalish(in_both, mappings):
    """Return string for triples that matched.

    Although triples "match", they may not come from the same orginal
    triples so do inverse mapping for each source graph.

    WARNING - uses _nt_row() from rdflib.plugins.serializers.nt which is
    a bit naughty. However, to avoid this it is necessary either to build
    a graph for each triple to be written, or to write another copy of
    the serialization code.
    """
    s = ''
    for triple in in_both:
        triple_in_first = from_bnodes_triple(triple, mappings[0])
        triple_in_second = from_bnodes_triple(triple, mappings[1])
        if (triple_in_first == triple_in_second):
            s += '== ' + _nt_row(triple_in_first)
        else:
            s += '=< ' + _nt_row(triple_in_first)
            s += '=> ' + _nt_row(triple_in_second)
    return s


def to_bnodes(graph, pattern):
    """Convert any URIs in graph matching pattern to bnodes.

    Parameters:
        graph -- input graph to map
        pattern -- a regex to be used with regex.search() on the string
            representations of subject and predicate terms in graph
    Returns:
        new_graph -- modified graph
        subs -- number of term substitutions made
        mapping -- dict with mapping of original URIRef -> BNode
    """
    new_graph = Graph()
    mapping = dict()
    regex = re.compile(pattern)
    logging.debug("Looking for %s in graph" % (str(regex)))
    subs = 0
    for s, p, o in graph:
        if (isinstance(s, URIRef)):
            if (s in mapping):
                s = mapping[s]
                subs += 1
            elif (regex.search(str(s))):
                mapping[s] = BNode()
                s = mapping[s]
                subs += 1
        if (isinstance(o, URIRef)):
            logging.debug("node %s %s %s" % (str(s), str(p), str(o)))
            if (o in mapping):
                o = mapping[o]
                subs += 1
            elif (regex.search(str(o))):
                mapping[o] = BNode()
                o = mapping[o]
                subs += 1
        new_graph.add((s, p, o))
    logging.debug("mapping: %s, made %d subs" % (str(mapping), subs))
    return new_graph, subs, mapping


def from_bnodes_triple(triple, mapping):
    """Convert bnodes back to original URI if listed in mapping, for one triple.

    Parameters:
      triple -- a single triple as tuple (s, p, o)
      mapping -- a dictionary that maps from the current node values
    to the desired ones.

    Returns:
      (s, p, o) -- tuple of possibly modified triple
    """
    s, p, o = triple
    if (s in mapping):
        s = mapping[s]
    if (o in mapping):
        o = mapping[o]
    return((s, p, o))


def from_bnodes(graph, mapping):
    """Convert bnodes back to original URI if listed in mapping."""
    new_graph = Graph()
    for triple in graph:
        new_graph.add(from_bnodes_triple(triple, mapping))
    return new_graph



"""Setup for rdiffb."""
from setuptools import setup, Command
import os
# setuptools used instead of distutils.core so that
# dependencies can be handled automatically

# Extract version number
import re
VERSIONFILE = "rdiffb/__init__.py"
verfilestr = open(VERSIONFILE, "rt").read()
match = re.search(
    r"^__version__ = '(\d\.\d.\d+(\.\d+)?)'",
    verfilestr,
    re.MULTILINE)
if match:
    version = match.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE))


class Coverage(Command):
    """Class to allow coverage run from setup."""

    description = "run coverage"
    user_options = []

    def initialize_options(self):
        """Empty initialize_options."""
        pass

    def finalize_options(self):
        """Empty finalize_options."""
        pass

    def run(self):
        """Run coverage program."""
        os.system("coverage run --source=rdiffb setup.py test")
        os.system("coverage report")
        os.system("coverage html")
        print("See htmlcov/index.html for details.")


setup(
    name='rdiffb',
    version=version,
    packages=['rdiffb'],
    scripts=['rdfdiffb.py'],
    classifiers=["Development Status :: 4 - Beta",
                 "Intended Audience :: Developers",
                 "License :: OSI Approved :: Apache Software License",
                 "Operating System :: OS Independent",
                 "Programming Language :: Python",
                 "Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3.6",
                 "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: 3.8",
                 "Topic :: Software Development :: Libraries :: Python Modules"],
    author='Simeon Warner',
    author_email='simeon.warner@cornell.edu',
    description='rdiffb',
    long_description=open('README').read(),
    url='http://github.com/zimeon/rdiffb',
    install_requires=[
        "rdflib>=4.2.0",
        "testfixtures"
    ],
    test_suite="tests",
    cmdclass={
        'coverage': Coverage,
    },
)
